import os
import subprocess
from pathlib import Path
from typing import Annotated

import typer

app = typer.Typer(no_args_is_help=True, help="CLI for managing jobs.")


@app.command(help="Start dagster server.")
def server(
    target: Annotated[
        Path,
        typer.Argument(
            help="Path that contains the scripts to run.",
        ),
    ] = Path("scripts"),
):
    repos_dir = Path(__file__).parent / "repos"

    dependencies = ["dagster-webserver", "uv"]

    command = ["uvx", "--with", ",".join(dependencies), "dagster", "dev"]

    for repo in repos_dir.glob("*.py"):
        command.extend(["-f", str(repo)])

    envs = os.environ.copy()
    envs["GYJD_SCRIPTS_PATH"] = str(target.absolute())

    process = subprocess.run(command, stdout=None, stderr=None, text=True, env=envs)

    print(f"Exit Code: {process.returncode}")


@app.command(help="Create python script job.")
def create_script(
    name: Annotated[str, typer.Argument(help="Name of the script.")],
    python_version: Annotated[str, typer.Option(help="Python version to use.", prompt=True)] = "3.11",
):
    print("Deleting user: Hiro Hamada")



import concurrent.futures
import functools
import itertools
import time
from collections.abc import Callable, Generator
from typing import Any, Literal

from gyjd.exceptions import (
    GYJDException,
    GYJDFailFastException,
    GYJDMultipleException,
    GYJDValueError,
)


class GYJDCallable:
    def __init__(
        self,
        func: Callable,
        return_exception_on_fail: bool = False,
        retry_attempts=-0,
        retry_delay=0,
        retry_max_delay=None,
        retry_backoff=1,
        retry_on_exceptions=(Exception,),
    ):
        self._func = func
        self._retry_attempts = retry_attempts
        self._retry_delay = retry_delay
        self._retry_max_delay = retry_max_delay
        self._retry_backoff = retry_backoff
        self._retry_on_exceptions = retry_on_exceptions
        self._return_exception_on_fail = return_exception_on_fail

    def __call__(self, *args, **kwargs):
        return self._attempt_call(*args, **kwargs)

    def _attempt_call(self, *args, **kwargs):
        raised_exceptions = []
        attempts = self._retry_attempts + 1
        delay = self._retry_delay

        while attempts >= 0:
            try:
                return self._func(*args, **kwargs)
            except GYJDFailFastException:
                return
            except self._retry_on_exceptions as e:
                raised_exceptions.append(e)

            attempts -= 1
            if not attempts:
                prepared_exception = GYJDMultipleException(raised_exceptions)
                if self._return_exception_on_fail:
                    return prepared_exception
                raise prepared_exception

            time.sleep(delay)
            delay *= self._retry_backoff
            if self._retry_max_delay is not None:
                delay = min(delay, self._retry_max_delay)

        raise GYJDException("This should never happen")

    def _call_with_parameters(self, parameters: dict[str, Any]) -> Any:
        return self.__call__(**parameters)

    def partial(self, *args, **kwargs) -> "GYJDCallable":
        return self._recreate(func=functools.partial(self._func, *args, **kwargs))

    def _recreate(self, **new_kwargs) -> "GYJDCallable":
        new_args = {
            "func": self._func,
            "retry_attempts": self._retry_attempts,
            "retry_delay": self._retry_delay,
            "retry_max_delay": self._retry_max_delay,
            "retry_backoff": self._retry_backoff,
            "retry_on_exceptions": self._retry_on_exceptions,
            "return_exception_on_fail": self._return_exception_on_fail,
            **new_kwargs,
        }

        return self.__class__(**new_args)

    def expand(
        self,
        parameters: dict[str, list[Any]] | list[dict[str, Any]],
        *,
        max_workers: int | None = None,
        strategy: Literal[
            "sequential",
            "thread_map",
            "thread_as_completed",
            "process_map",
            "process_as_completed",
        ] = "sequential",
    ) -> Generator[Any, None, None]:
        if isinstance(parameters, dict):
            generated_parameters = (
                dict(zip(parameters.keys(), comb)) for comb in itertools.product(*parameters.values())
            )
        else:
            generated_parameters = parameters

        if strategy == "sequential":
            for generated_parameter in generated_parameters:
                yield self._call_with_parameters(generated_parameter)
            return

        executor_type, execution_mode = strategy.split("_", 1)

        executor_cls = {
            "thread": concurrent.futures.ThreadPoolExecutor,
            "process": concurrent.futures.ProcessPoolExecutor,
        }[executor_type]

        with executor_cls(max_workers=max_workers) as executor:
            if execution_mode == "map":
                for result in executor.map(self._call_with_parameters, generated_parameters):
                    yield result
            elif execution_mode == "as_completed":
                tasks = (
                    executor.submit(self._call_with_parameters, generated_parameter)
                    for generated_parameter in generated_parameters
                )
                for future in concurrent.futures.as_completed(tasks):
                    yield future.result()
            else:
                raise GYJDValueError(f"Invalid strategy: {strategy}")



import os
import shutil
import subprocess
from pathlib import Path
from typing import Annotated

import typer

from gyjd.cli.apps.jobs.app import app as jobs_app

app = typer.Typer(no_args_is_help=True)
app.add_typer(jobs_app, name="jobs", help="CLI for managing jobs.")


@app.command(name="compile", help="Compile a Python file to an executable.", no_args_is_help=True)
def compile(
    filename: Annotated[
        Path,
        typer.Option(help="Python file to compile."),
    ],
):
    output_dir = "dist"

    commnad = [
        "uvx",
        "nuitka",
        "--follow-imports",
        "--onefile",
        f"--output-dir={output_dir}",
        "--assume-yes-for-downloads",
        str(filename),
    ]

    subprocess.run(commnad, stdout=None, stderr=None, text=True)

    for entry in os.listdir(output_dir):
        entry_uri = os.path.join(output_dir, entry)
        if not os.path.isfile(entry_uri):
            shutil.rmtree(entry_uri)


if __name__ == "__main__":
    app()



import logging

from gyjd.config import LoggerConfig
from gyjd.core.simple_injector import inject_dependencies


class GYJDLogger(logging.Logger): ...


@inject_dependencies
def get_default_logger(config: LoggerConfig):
    logger = logging.getLogger(config.name)
    if not logger.handlers and config.default_to_console:
        handler = logging.StreamHandler()
        formatter = logging.Formatter(config.formatter)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(getattr(logging, config.level.upper()))
    return logger



class GYJDException(Exception): ...


class GYJDMultipleException(GYJDException):
    def __init__(self, exceptions: list[Exception]):
        self.exceptions = exceptions


class GYJDFailFastException(GYJDException): ...


class GYJDValueError(GYJDException, ValueError): ...



import asyncio
import sqlite3
from contextlib import asynccontextmanager, contextmanager
from datetime import datetime
from typing import Literal


class SQLiteConnection:
    _MAINTENANCE_TABLE = "maintenance_metadata"

    def __init__(self, conn_str: str):
        self.conn = sqlite3.connect(conn_str, check_same_thread=False)
        self.conn.execute("PRAGMA foreign_keys = ON")
        self.conn.execute("PRAGMA journal_mode = WAL;")
        self._ensure_maintenance_metadata()
        self.auto_maintenance()
        self._lock = asyncio.Lock()

    def _ensure_maintenance_metadata(self):
        sql = f"CREATE TABLE IF NOT EXISTS {self._MAINTENANCE_TABLE} (event_name VARCHAR(36) PRIMARY KEY, last_event_datetime DATETIME);"
        self.conn.execute(sql)

    def _register_event(self, event_name):
        sql = f"INSERT OR REPLACE INTO {self._MAINTENANCE_TABLE} (event_name, last_event_datetime) VALUES (?, ?);"
        with self.cursor() as c:
            c.execute(sql, (event_name, datetime.utcnow().isoformat()))

    def _get_last_event_datetime(self, event_name) -> datetime | None:
        sql = f"SELECT last_event_datetime FROM {self._MAINTENANCE_TABLE} WHERE event_name = ?;"
        with self.cursor() as c:
            c.execute(sql, (event_name,))
            result = c.fetchone()
            return datetime.fromisoformat(result[0]) if result else None

    def vacuum(self):
        self.conn.execute("VACUUM;")
        self._register_event("vacuum")

    def checkpoint(self, mode: Literal["passive", "full", "restart", "truncate"] = "full"):
        self.conn.execute(f"PRAGMA wal_checkpoint({mode});")
        self._register_event("checkpoint")

    def auto_maintenance(self):
        last_vacuum = self._get_last_event_datetime("vacuum")
        if last_vacuum is None or (datetime.utcnow() - last_vacuum).days > 7:
            self.vacuum()

        last_checkpoint = self._get_last_event_datetime("checkpoint")
        if last_checkpoint is None or (datetime.utcnow() - last_checkpoint).days > 1:
            self.checkpoint()

    @contextmanager
    def cursor(self):
        c = self.conn.cursor()
        try:
            yield c
        except Exception:
            self.conn.rollback()
            raise
        finally:
            c.close()
            self.conn.commit()

    @asynccontextmanager
    async def async_cursor(self):
        async with self._lock:
            with self.cursor() as c:
                yield c



import time
from dataclasses import dataclass, field
from datetime import datetime
from os import getenv
from uuid import uuid4


@dataclass
class LoggerConfig:
    name: str = "gyjd"
    level: str = field(default_factory=lambda: getenv("LOG_LEVEL", "INFO"))
    default_to_console: bool = True
    formatter: str = field(
        default_factory=lambda: getenv("LOG_FORMAT", "%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )


@dataclass
class Config:
    logger: LoggerConfig
    node_id: str = field(default_factory=lambda: f"{hex(int(time.time()*1000))[2:]}-{uuid4()}")
    start_time: datetime = field(default_factory=datetime.now)



import os
from pathlib import Path
from typing import Literal

from gyjd.database.sqlite_connection import SQLiteConnection


class ConnectionFactory:
    @classmethod
    def create_connection(cls, db_name: Literal["event_bus"]) -> SQLiteConnection:
        location = Path.home() / "gyjd" / "database" / db_name / f"{db_name}.db"
        os.makedirs(location.parent, exist_ok=True)
        conn = SQLiteConnection(str(location.absolute()))
        getattr(cls, f"_create_{db_name}_schema")(conn)
        return conn

    @classmethod
    def _create_event_bus_schema(cls, conn: SQLiteConnection) -> None:
        sql = """
            CREATE TABLE IF NOT EXISTS events (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                event_type TEXT NOT NULL,
                payload TEXT NOT NULL,
                processed INTEGER NOT NULL DEFAULT 0,
                created_at DATETIME NOT NULL,
                processed_at DATETIME
            );
            CREATE TABLE IF NOT EXISTS subscribers (
                task_name TEXT PRIMARY KEY,
                event_types TEXT NOT NULL,
                function_path TEXT NOT NULL,
                mode TEXT NOT NULL DEFAULT 'any',
                created_at DATETIME NOT NULL,
                max_attempts INTEGER NOT NULL DEFAULT 1,
                retry_delay INTEGER NOT NULL DEFAULT 30,
                concurrency_limit INTEGER NOT NULL DEFAULT 8
            );
            CREATE TABLE IF NOT EXISTS tasks (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                subscriber_task_name TEXT NOT NULL,
                status TEXT NOT NULL DEFAULT 'pending',
                attempt_count INTEGER NOT NULL DEFAULT 0,
                handled_events TEXT NOT NULL,
                parameters TEXT NOT NULL,
                last_attempt_at DATETIME,
                created_at DATETIME NOT NULL,
                completed_at DATETIME,
                scheduled_at DATETIME NOT NULL,
                FOREIGN KEY(subscriber_task_name) REFERENCES subscribers(task_name)
            );
        """
        conn.conn.executescript(sql)



import os
import tomllib
from dataclasses import fields, is_dataclass
from datetime import date
from functools import cache
from typing import Any, Type, TypeVar

T = TypeVar("T")


def cast_value(value: Any, expected_type: Type) -> Any:
    if isinstance(value, str) and value.startswith("$Env:"):
        env_part = value.removeprefix("$Env:")

        if "|" in env_part:
            env_var_name, default_value = env_part.split("|", 1)
        else:
            env_var_name, default_value = env_part, None

        value = os.getenv(env_var_name, default_value)

        if value is None:
            raise ValueError(f"Environment variable '{env_var_name}' is not set and no default value was provided")

    if isinstance(value, expected_type):
        return value

    try:
        if expected_type is date:
            return date.fromisoformat(value)
        elif expected_type is bool:
            if isinstance(value, str) and value.lower() in {"true", "false"}:
                return value.lower() == "true"
            return bool(value)
        return expected_type(value)
    except (ValueError, TypeError):
        raise ValueError(f"Cannot cast value '{value}' of type '{type(value).__name__}' to '{expected_type}'")


def load_data_to_dataclass(data: dict, dataclass_type: type[T]) -> T:
    field_values = {}

    for field_info in fields(dataclass_type):
        field_name = field_info.name
        field_type = field_info.type
        value = data.get(field_name)

        if value is not None:
            if is_dataclass(field_type):
                field_values[field_name] = load_data_to_dataclass(value, field_type)
            else:
                field_values[field_name] = cast_value(value, field_type)

    return dataclass_type(**field_values)


@cache
def load_file(filepath: str) -> dict:
    if not os.path.exists(filepath):
        raise FileNotFoundError(f"Config file '{filepath}' not found")

    with open(filepath, "rb") as file:
        data = tomllib.load(file)

    return data


def load_config_from_toml_file(filepath: str, subtree: list[str] | str | None = None) -> dict:
    if subtree is not None:
        if isinstance(subtree, str):
            subtree = [subtree]
    else:
        subtree = []

    data = load_file(filepath)

    for item in subtree:
        if item not in data:
            raise KeyError(f"Path '{item}' not found in config file '{filepath}'")
        data = data[item]

    return data


def load_config_file(
    config_type: type[T],
    filepath: str,
    allow_if_file_not_found: bool = False,
    subtree: list[str] | str | None = None,
):
    try:
        data = load_config_from_toml_file(filepath, subtree)
        return load_data_to_dataclass(data, config_type)
    except FileNotFoundError:
        if not allow_if_file_not_found:
            raise
        return config_type()



import functools
import inspect
from functools import wraps
from typing import Callable, Literal, Type

_DEPENDENCIES_REGISTER: dict[Type, "DependencyHandler"] = {}
IF_EXISTS_TYPE = Literal["raise", "skip", "overwrite"]


class DependencyHandler:
    def __init__(self, instance_builder: Callable, reuse_times: int = -1):
        self._instance_generator = self._build_instance_generator(
            instance_builder=instance_builder,
            reuse_times=reuse_times,
        )

    @staticmethod
    def _build_instance_generator(instance_builder: Callable, reuse_times: int):
        if reuse_times == -1:
            instance = instance_builder()
            while True:
                yield instance

        while True:
            instance = instance_builder()
            if reuse_times == 0:
                yield instance
            else:
                for _ in range(reuse_times):
                    yield instance
            del instance

    def get(self):
        return next(self._instance_generator)


def inject_dependencies(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        func_signature = inspect.signature(func)
        bound_arguments = func_signature.bind_partial(*args, **kwargs)

        for param_name, param in func_signature.parameters.items():
            param_type = param.annotation

            if param_type is not param.empty and param_name not in bound_arguments.arguments:
                found_dependency = _DEPENDENCIES_REGISTER.get(param_type)
                if found_dependency and param_name not in kwargs:
                    kwargs[param_name] = found_dependency.get()

        return func(*args, **kwargs)

    return wrapper


def register_dependency(
    func=None,
    reuse_times: int = -1,
    cls: type | None = None,
    if_exists: IF_EXISTS_TYPE = "raise",
):
    if func is None:
        return functools.partial(register_dependency, reuse_times=reuse_times, cls=cls)

    if cls is None:
        if inspect.isclass(func):
            cls = func
        else:
            cls = func.__annotations__.get("return")
            if cls is None:
                raise ValueError("No return type annotation found, please provide a class type")

    if if_exists == "raise" and cls in _DEPENDENCIES_REGISTER:
        raise ValueError(f"Dependency of type {cls} already registered")

    if if_exists == "skip" and cls in _DEPENDENCIES_REGISTER:
        return func

    _DEPENDENCIES_REGISTER[cls] = DependencyHandler(
        inject_dependencies(func),
        reuse_times=reuse_times,
    )

    return func


def get_registered_dependencies() -> set[Type]:
    return set(_DEPENDENCIES_REGISTER)


def clear_registered_dependencies():
    _DEPENDENCIES_REGISTER.clear()



import os
import subprocess
import sys
from dataclasses import dataclass, field
from pathlib import Path

import toml
from dagster import Field, OpExecutionContext, ScheduleDefinition, Shape, job, op, repository


@dataclass
class ScriptConfig:
    script_path: str
    script_name: str
    python_version: str
    dependencies: list[str] = field(default_factory=list)


@op(
    config_schema=Shape(
        {
            "script_path": Field(str, description="Caminho para o script Python"),
            "python_version": Field(str, description="Versão do Python"),
            "dependencies": Field([str], description="Dependências do script"),
        }
    )
)
def run_python_script(context: OpExecutionContext):
    script_path = context.op_config["script_path"]
    python_version = context.op_config["python_version"]
    dependencies = context.op_config["dependencies"]

    context.log.info(f"Starting script: {script_path}")

    command = [sys.executable, "-m", "uv", "run", "--no-project"]

    if python_version:
        command.extend(["--python", python_version])

    if dependencies:
        command.extend(["--with", ",".join(dependencies)])

    command.extend(["--script", str(script_path)])

    envs = os.environ.copy()

    logger_prefix = "gyjd-jobs -"

    envs["LOG_FORMATTER"] = f"{logger_prefix} - %(levelname)s - %(message)s"

    process = subprocess.Popen(command, stdout=None, stderr=None, text=True, env=envs)

    exit_code = process.wait()
    if exit_code != 0:
        raise Exception(f"Script finished with exit code: {exit_code}")

    context.log.info(f"Script finished with exit code: {exit_code}")


# Função para criar um job para cada script
def create_job_for_script(config: ScriptConfig):
    @job(name=config.script_name)
    def dynamic_job():
        run_python_script.configured(
            {
                "script_path": config.script_path,
                "python_version": config.python_version,
                "dependencies": config.dependencies,
            },
            name=f"{config.script_name}_op",
        )()

    return dynamic_job


def generate_definitions(scripts_path: Path):
    for config_path in scripts_path.glob("**/*.toml"):
        with open(config_path) as f:
            try:
                script_config = toml.load(f).get("gyjd", {}).get("job", {})
            except toml.TomlDecodeError:
                continue

        if "script" not in script_config:
            continue

        script_path = (scripts_path / config_path).parent / script_config["script"]

        config = ScriptConfig(
            script_path=str(script_path),
            script_name=script_config.get("name", script_path.stem),
            python_version=script_config.get("python_version", f"{sys.version_info.major}.{sys.version_info.minor}"),
            dependencies=script_config.get("dependencies", []),
        )

        job = create_job_for_script(config=config)

        yield job

        schedules = script_config.get("schedule", {})

        for schedule_name, schedule_config in schedules.get("cron", {}).items():
            cron_expression = schedule_config.get("expression")
            if not cron_expression:
                continue

            cron_timezone = schedule_config.get("timezone", "UTC")

            yield ScheduleDefinition(
                job=job,
                cron_schedule=cron_expression,
                execution_timezone=cron_timezone,
                name=f"{config.script_name}_{schedule_name}_schedule",
            )


@repository(
    name="scripts_repository",
    description="Repositório com jobs e schedules para scripts Python em ambientes isolados utilizando uv coordenadamente",
)
def scripts_repository():
    scripts_path = Path(os.environ["GYJD_SCRIPTS_PATH"])
    return list(generate_definitions(scripts_path=scripts_path))



import logging
from collections.abc import Callable
from dataclasses import fields, is_dataclass
from functools import partial

from gyjd.config import LoggerConfig
from gyjd.core.config_loader import load_config_file
from gyjd.core.gyjd_callable import GYJDCallable
from gyjd.core.logger import GYJDLogger, get_default_logger
from gyjd.core.simple_injector import clear_registered_dependencies, inject_dependencies, register_dependency


def setup_defaults(clear_dependencies: bool = False):
    """
    Register default dependencies:
    - GYJDLogger
    - logging.Logger
    - LoggerConfig

    If clear_dependencies is True, clear all registered dependencies before registering the default ones.
    """

    if clear_dependencies:
        clear_registered_dependencies()

    register_dependency(get_default_logger, cls=GYJDLogger, reuse_times=-1, if_exists="skip")
    register_dependency(get_default_logger, cls=logging.Logger, reuse_times=-1, if_exists="skip")
    register_dependency(LoggerConfig, reuse_times=-1, if_exists="skip")


class gyjd:
    register_dependency = partial(register_dependency, if_exists="overwrite")

    def __new__(
        cls,
        func: Callable | None = None,
        *,
        return_exception_on_fail: bool = False,
        retry_attempts=-0,
        retry_delay=0,
        retry_max_delay=None,
        retry_backoff=1,
        retry_on_exceptions=(Exception,),
    ) -> GYJDCallable:
        if func is None:
            wrapper = partial(
                gyjd,
                return_exception_on_fail=return_exception_on_fail,
                retry_attempts=retry_attempts,
                retry_delay=retry_delay,
                retry_max_delay=retry_max_delay,
                retry_backoff=retry_backoff,
                retry_on_exceptions=retry_on_exceptions,
            )
            return wrapper

        return GYJDCallable(
            func=inject_dependencies(func),
            return_exception_on_fail=return_exception_on_fail,
            retry_attempts=retry_attempts,
            retry_delay=retry_delay,
            retry_max_delay=retry_max_delay,
            retry_backoff=retry_backoff,
            retry_on_exceptions=retry_on_exceptions,
        )

    @classmethod
    def _collect_children_config(cls, dataclass_type: type, subtree: str = ""):
        for field in fields(dataclass_type):
            full_tree = f"{subtree}.{field.name}" if subtree else field.name
            if is_dataclass(field.type):
                yield full_tree, field.type
                yield from cls._collect_children_config(field.type, full_tree)

    @classmethod
    def register_config_file(
        cls,
        *,
        config_type: type,
        filepath: str,
        allow_if_file_not_found: bool = False,
        subtree: str = "",
    ) -> None:
        subtree = subtree.strip(".")

        base_loader = partial(
            load_config_file,
            filepath=filepath,
            allow_if_file_not_found=allow_if_file_not_found,
        )

        register_dependency(
            partial(
                base_loader,
                config_type=config_type,
                subtree=subtree.split("."),
            ),
            cls=config_type,
            reuse_times=-1,
        )

        for child_subtree, child_type in cls._collect_children_config(config_type):
            register_dependency(
                partial(
                    base_loader,
                    config_type=child_type,
                    subtree=child_subtree.split("."),
                ),
                cls=child_type,
                reuse_times=-1,
                if_exists="overwrite",
            )


setup_defaults()
__all__ = ["gyjd"]



import asyncio
import importlib
import json
import logging
from datetime import datetime, timedelta
from typing import List, Literal, TypedDict

from gyjd.database.connection_factory import ConnectionFactory

DEFAULT_RETRY_DELAY = 30

logger = logging.getLogger("gyjd")


class MappedEventDict(TypedDict):
    event_id: str
    event_name: str
    payload: dict
    event_date: str


MappedEvent = List[MappedEventDict]


class EventBus:
    def __init__(self, polling_interval=10):
        self._conn = ConnectionFactory.create_connection("event_bus")
        self.polling_interval = polling_interval

    def add_event(self, event_type: str, payload: dict):
        self._conn.conn.execute(
            "INSERT INTO events (event_type, payload, created_at) VALUES (?, ?, CURRENT_TIMESTAMP)",
            (event_type, json.dumps(payload)),
        )

    def subscribe(
        self,
        event_types: List[str],
        function_path: str,
        task_name: str,
        mode: Literal["any", "batch"] = "any",
        max_attempts: int = 1,
        retry_delay: int = DEFAULT_RETRY_DELAY,
        concurrency_limit: int | None = None,
    ):
        if concurrency_limit is None:
            concurrency_limit = 8 if mode == "any" else 1

        if mode == "batch" and concurrency_limit > 1:
            raise ValueError("concurrency_limit must be 1 for batch mode")

        with self._conn.cursor() as c:
            sql = """
                INSERT OR REPLACE INTO subscribers (task_name, event_types, function_path, mode, max_attempts, retry_delay, concurrency_limit, created_at)
                VALUES (?, ?, ?, ?, ?, ?, ?, CURRENT_TIMESTAMP)
            """
            c.execute(
                sql,
                (
                    task_name,
                    json.dumps(event_types),
                    function_path,
                    mode,
                    max_attempts,
                    retry_delay,
                    concurrency_limit,
                ),
            )

    async def process_events(self) -> bool:
        logger.info("Looking for events to process")
        async with self._conn.async_cursor() as c:
            c.execute("SELECT min(id) min_id, max(id) max_id FROM events WHERE processed = 0")
            min_id, max_id = c.fetchone()

            if not (min_id and max_id):
                logger.info("No events to process")
                return False

            c.execute(
                """
                    INSERT INTO tasks (subscriber_task_name, handled_events, parameters, created_at, scheduled_at)
                    WITH subs as (
                        SELECT task_name, value as event_type
                        FROM subscribers, json_each(event_types)
                        WHERE mode = 'any'
                    )
                    SELECT
                        task_name as subscriber_task_name,
                        '[' || e.id || ']' as handled_events,
                        '[' || payload || ']' as parameters,
                        CURRENT_TIMESTAMP as created_at,
                        CURRENT_TIMESTAMP as scheduled_at
                    FROM events e join subs using(event_type)
                    WHERE e.processed = 0 AND e.id BETWEEN ? AND ?
                """,
                (min_id, max_id),
            )

            c.execute(
                """
                    INSERT INTO tasks (subscriber_task_name, handled_events, parameters, created_at, scheduled_at)
                    WITH subs as (
                        SELECT task_name, value as event_type
                        FROM subscribers, json_each(event_types)
                        WHERE mode = 'batch'
                    )
                    SELECT
                        task_name as subscriber_task_name,
                        '[' || GROUP_CONCAT(e.id, ', ') || ']' as handled_events,
                        '[' || GROUP_CONCAT(e.payload, ', ') || ']' as parameters,
                        CURRENT_TIMESTAMP as created_at,
                        CURRENT_TIMESTAMP as scheduled_at
                    FROM events e join subs using(event_type)
                    WHERE e.processed = 0 AND e.id BETWEEN ? AND ?
                    GROUP BY 1
                """,
                (min_id, max_id),
            )

            c.execute(
                "UPDATE events SET processed = 1, processed_at = CURRENT_TIMESTAMP WHERE processed = 0 AND id BETWEEN ? AND ?",
                (min_id, max_id),
            )

            return True

    @classmethod
    def _load_function(cls, function_path: str):
        parts = function_path.split(".")
        mod_name = ".".join(parts[:-1])
        func_name = parts[-1]
        module = importlib.import_module(mod_name)
        func = getattr(module, func_name)
        return func

    async def _run_task(self, task):
        (task_id, subscriber_task_name, attempt_count, parameters) = task

        async with self._conn.async_cursor() as c:
            c.execute(
                "SELECT function_path, max_attempts, retry_delay FROM subscribers WHERE task_name = ?",
                (subscriber_task_name,),
            )
            row = c.fetchone()
            if not row:
                c.execute("UPDATE tasks SET status = 'failed' WHERE id = ?", (task_id,))
                return

            function_path, max_attempts, retry_delay = row

        async with self._conn.async_cursor() as c:
            c.execute(
                "UPDATE tasks SET attempt_count = attempt_count + 1, last_attempt_at = CURRENT_TIMESTAMP WHERE id = ?",
                (task_id,),
            )

        try:
            func = self._load_function(function_path)
            parameters = json.loads(parameters)
        except Exception:
            async with self._conn.async_cursor() as c:
                c.execute("UPDATE tasks SET status = 'failed' WHERE id = ?", (task_id,))
            return

        try:
            if asyncio.iscoroutinefunction(func):
                await func(parameters)
            else:
                loop = asyncio.get_running_loop()
                await loop.run_in_executor(None, func, parameters)

            async with self._conn.async_cursor() as c:
                c.execute(
                    "UPDATE tasks SET status = 'done', completed_at = CURRENT_TIMESTAMP WHERE id = ?",
                    (task_id,),
                )

        except Exception as e:
            print(e)
            async with self._conn.async_cursor() as c:
                c.execute("SELECT attempt_count FROM tasks WHERE id = ?", (task_id,))
                attempt_count = c.fetchone()[0]
                if attempt_count < max_attempts:
                    new_time = (datetime.utcnow() + timedelta(seconds=retry_delay)).isoformat()
                    c.execute("UPDATE tasks SET status = 'pending', scheduled_at = ? WHERE id = ?", (new_time, task_id))
                else:
                    c.execute("UPDATE tasks SET status = 'failed' WHERE id = ?", (task_id,))

    async def run_tasks(self) -> bool:
        logger.info("Looking for tasks to run")
        sql = """
            WITH bs AS (
                SELECT
                    t.id,
                    t.subscriber_task_name,
                    t.attempt_count,
                    t.parameters,
                    t.scheduled_at,
                    s.concurrency_limit,
                    row_number() OVER (PARTITION BY t.subscriber_task_name ORDER BY t.scheduled_at) AS rn
                FROM
                    tasks t
                JOIN subscribers s ON
                    t.subscriber_task_name = s.task_name
                WHERE
                    t.scheduled_at <= CURRENT_TIMESTAMP AND
                    t.status = 'pending'
            )
            SELECT
                id,
                subscriber_task_name,
                attempt_count,
                parameters
            FROM bs
            WHERE rn <= concurrency_limit
            ORDER BY scheduled_at
            LIMIT 16
        """

        async with self._conn.async_cursor() as c:
            c.execute(sql)
            tasks = c.fetchall()

        if not tasks:
            logger.info("No tasks to run")
            return False

        await asyncio.gather(*(self._run_task(t) for t in tasks))

        return True

    async def run(self):
        while await self.run_tasks() or await self.process_events():
            pass

    async def run_forever(self):
        logger.info("Event bus started")
        while True:
            await self.run()


event_bus = EventBus()


def subscribe(
    event_types: List[str],
    task_name: str | None = None,
    mode: Literal["any", "batch"] = "any",
    max_attempts: int = 1,
    retry_delay: int = DEFAULT_RETRY_DELAY,
    concurrency_limit: int = 8,
):
    def decorator(func):
        module_name: str = func.__module__
        func_name: str = func.__name__
        path = f"{module_name}.{func_name}"
        nonlocal task_name
        if task_name is None:
            task_name = func_name
        event_bus.subscribe(event_types, path, task_name, mode, max_attempts, retry_delay, concurrency_limit)
        return func

    return decorator


def emmit(*, event_type: str, payload: dict):
    event_bus.add_event(event_type, payload)
